/*
Copyright 2020 Tommi Jauhiainen
Copyright 2021 University of Helsinki
Copyright 2021 Heidi Jauhiainen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
/*
 HeLI-OTS 1.1.
 
 If you use this program in producing scientific publications, please refer to:
 
 @inproceedings{jauhiainen-etal-2017-evaluation,
     title = "Evaluation of language identification methods using 285 languages",
     author = "Jauhiainen, Tommi  and
       Lind{\'e}n, Krister  and
       Jauhiainen, Heidi",
     booktitle = "Proceedings of the 21st Nordic Conference on Computational Linguistics",
     month = may,
     year = "2017",
     address = "Gothenburg, Sweden",
     publisher = "Association for Computational Linguistics",
     url = "https://www.aclweb.org/anthology/W17-0221",
     pages = "183--191",
 }
 
 Producing and publishing this software has been partly supported by The Finnish Research Impact Foundation Tandem Industry Academia -funding in cooperation with Lingsoft.
 */

import java.io.*;
import java.util.*;
//import org.apache.commons.lang.ArrayUtils;

class HeLI {
	
    private static TreeMap<String, TreeMap<String, Float>> gramDict;
    private static TreeMap<String, TreeMap<String, Float>> wordDict;
	private static List<String> languageList = new ArrayList<String>();
	
// The following values are the ones used in Jauhiainen et al. 2017.

	private static float usedmonos = (float) 0.0000005;
	private static float usedbis = (float) 0.0000005;
	private static float usedtris = (float) 0.0000005;
	private static float usedquads = (float) 0.0000005;
	private static float usedcinqs = (float) 0.0000005;
	private static float usedsexts = (float) 0.0000005;
	private static float usedwords = (float) 0.0000005;
	
	private static float penaltyValue = (float) 7.0;
	
	private static int maxNgram = 6;
    
    private static boolean useConfidence = false;
    private static boolean filterMode = false;
		
	public static void main(String[] args) {
        
        String allHelp = "HeLI off-the-shelf language identifier with language models for 200 languages (http://urn.fi/urn:nbn:fi:lb-2021062801). The program will read the <infile> and classify the language of each line as one of the 200 languages it knows and writes the results, one ISO 639-3 code per line, into file <outfile>. It can identify c. 3000 sentences per second using one core on a 2021 laptop and around 3 gigabytes of memory.\n" +
        
        "Producing and publishing this software has been partly supported by The Finnish Research Impact Foundation Tandem Industry Academia -funding in cooperation with Lingsoft.\n" +
        "You can give the name of the text file to be identified as the first argument and the name of a new file to write the results in as the second argument.\n\n"+
        "Usage: java -jar HeLI.jar <infile> <outfile>\n\n"+
        "You can use the -c option to make the program print a confidence score for the identification after each language code.\n\n"+
        "Usage: java -jar HeLI.jar -c <infile> <outfile>\n\n"+
        "If you omit both of the filenames, the program will read the standard input one line at a time and write the result to standard output.";
        
        if (args.length == 0) {
            filterMode = true;
        }
        else if (args[0].equals("-h")) {
            System.out.println(allHelp);
            System.exit(0);
        }
        else if (args[0].equals("-c") && args.length == 1) {
            filterMode = true;
            useConfidence = true;
        }
        else if (args.length == 1 || (args[0].equals("-c") && args.length == 2)) {
            System.out.println("You have to give the name of a new file to write the results in as the second argument.");
            System.out.println("Usage: java -jar HeLI.jar <infile> <outfile>");
            System.exit(0);
        }
        else if (args[0].equals("-c")) {
            useConfidence = true;
            String[] tempArgs = new String[] {args[1],args[2]};
            args = tempArgs;
        }
        
        String mysteryTexts = "";
        String resultsFile = "";
        File file;
        
        if (!filterMode) {
            mysteryTexts = args[0];
            resultsFile = args[1];
            
            file = new File(resultsFile);
            if (file.exists()) {
                System.out.println("The specified outfile already exists, define a new file.");
                System.exit(0);
            }
        }
        
        InputStream in = HeLI.class.getResourceAsStream("languagelist");
        
		BufferedReader reader = null;
		
		try {
            reader = new BufferedReader(new InputStreamReader(in));
			String text = null;
			while ((text = reader.readLine()) != null) {
				languageList.add(text);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				if (reader != null) {
					reader.close();
				}
			} catch (IOException e) {
			}
		}
		        
        gramDict = new TreeMap<>();
        wordDict = new TreeMap<>();
		
		ListIterator gramiterator = languageList.listIterator();
		while(gramiterator.hasNext()) {
			Object element = gramiterator.next();
			String languageCode = (String) element;

			loadModel(usedmonos, languageCode, "mono");
			loadModel(usedbis, languageCode, "bi");
			loadModel(usedtris, languageCode, "tri");
			loadModel(usedquads, languageCode, "quad");
			loadModel(usedcinqs, languageCode, "cinq");
			loadModel(usedsexts, languageCode, "sext");
			loadModel(usedwords, languageCode, "word");
		}
        
        if (!filterMode) {
            file = new File(mysteryTexts);
            
            reader = null;
            try {
                reader = new BufferedReader(new FileReader(file));
                String line = "";
                
                file = new File(resultsFile);
                file.createNewFile();
                BufferedWriter writer = null;
                try {
                    writer = new BufferedWriter(new FileWriter(file, true));
                }
                catch (Exception e) {
                    System.out.println("Error while creating writer: "+e.getMessage());
                }
                
                while ((line = reader.readLine()) != null) {
                    try {
                        writer.write(HeLI.identifyLanguage(line)+"\n");
                    }
                    catch (Exception e) {
                        System.out.println("Error while trying to write: "+e.getMessage());
                    }
                }
                
                try {
                    if (writer != null) {
                        writer.close();
                    }
                }
                catch (Exception e) {
                    System.out.println(e.getMessage());
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                try {
                    if (reader != null) {
                        reader.close();
                    }
                } catch (IOException e) {
                }
            }
        }
        else {
            Scanner heliLoop = new Scanner(System.in);
            while (heliLoop.hasNextLine()) {
                String line = heliLoop.nextLine();
                System.out.println(HeLI.identifyLanguage(line));
            }
        }
	}
	
	private static void loadModel(float usedFeatureRF, String languageCode, String modelType) {
        TreeMap<String, Float> tempDict;
		
        tempDict = new TreeMap<>();
	
        InputStream modelFile = null;
        InputStream modelLengthFile = null;
	
		if (modelType.equals("word")) {
            modelFile = HeLI.class.getResourceAsStream("LanguageModels/" + languageCode + ".wordcount");
			modelLengthFile = HeLI.class.getResourceAsStream("LanguageModels/" + languageCode + ".words.count");
		}
		else {
			modelFile = HeLI.class.getResourceAsStream("LanguageModels/" + languageCode + "-" + modelType + ".X");
			modelLengthFile = HeLI.class.getResourceAsStream("LanguageModels/" + languageCode + "-" + modelType + ".count");
		}
	
		float totalFeatureNumber = 0;
		float langamount = 0;
	
		BufferedReader reader = null;
		try {
            reader = new BufferedReader(new InputStreamReader(modelLengthFile));
			String text = null;
			while ((text = reader.readLine()) != null) {
				totalFeatureNumber = Float.parseFloat(text);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				if (reader != null) {
					reader.close();
				}
			} catch (IOException e) {
			}
		}

		reader = null;
		try {
            reader = new BufferedReader(new InputStreamReader(modelFile));
			String text = null;
			while ((text = reader.readLine()) != null) {
				String gram = text;
				
				if (modelType.equals("word")) {
					gram = gram.replaceAll(".*[1-90]", "");
					gram = gram.replaceAll("$", " ");
				}
				else {
					gram = text.replaceAll(".*[1-90]=", "");
				}

				int amount = Integer.parseInt(text.replaceAll("[^1-90]", ""));
				
				if (amount/totalFeatureNumber > usedFeatureRF) {
					tempDict.put(gram, (float) amount);
					langamount = langamount + (float) amount;
				}
				else {
					break;
				}				
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				if (reader != null) {
					reader.close();
				}
			} catch (IOException e) {
			}
		}				

        for (Map.Entry<String,Float> entry : tempDict.entrySet()) {
			Float probability = (float) -Math.log10(entry.getValue() / langamount);
            TreeMap <String, Float> kiepro = new TreeMap<>();
   			if (modelType.equals("word")) {
                if (wordDict.containsKey(entry.getKey())) {
                    kiepro = wordDict.get(entry.getKey());
                }
                kiepro.put(languageCode,probability);
                wordDict.put(entry.getKey(), kiepro);
			}
			else {
                if (gramDict.containsKey(entry.getKey())) {
                    kiepro = gramDict.get(entry.getKey());
                }
                kiepro.put(languageCode,probability);
				gramDict.put(entry.getKey(), kiepro);
			}
		}
	}
    
	public static String identifyLanguage(String mysteryText) {
		mysteryText = mysteryText.toLowerCase();
		
		mysteryText = mysteryText.replaceAll("[^\\p{L}\\p{M}′'’´ʹािीुूृेैोौंँः् া ি ী ু ূ ৃ ে ৈ ো ৌ।্্্я̄\\u07A6\\u07A7\\u07A8\\u07A9\\u07AA\\u07AB\\u07AC\\u07AD\\u07AE\\u07AF\\u07B0\\u0A81\\u0A82\\u0A83\\u0ABC\\u0ABD\\u0ABE\\u0ABF\\u0AC0\\u0AC1\\u0AC2\\u0AC3\\u0AC4\\u0AC5\\u0AC6\\u0AC7\\u0AC8\\u0AC9\\u0ACA\\u0ACB\\u0ACC\\u0ACD\\u0AD0\\u0AE0\\u0AE1\\u0AE2\\u0AE3\\u0AE4\\u0AE5\\u0AE6\\u0AE7\\u0AE8\\u0AE9\\u0AEA\\u0AEB\\u0AEC\\u0AED\\u0AEE\\u0AEF\\u0AF0\\u0AF1]", " ");

		String mysteryText2 = "";
		int lastWasCJK = 0;
		int lastWasSpace = 0;
		int CJKcharacterAmount = 0;
		
		for (int charCounter = 0; charCounter < mysteryText.length(); charCounter++){
			char mysteryChar = mysteryText.charAt(charCounter);
			String mysteryCharSet;
			try {
				mysteryCharSet = Character.UnicodeBlock.of(mysteryChar).toString();
			}
			catch (Exception e) {
				return("xxx");
			}
			if (mysteryCharSet.startsWith("CJK")) {
				if (lastWasCJK == 0 && lastWasSpace == 0) {
					mysteryText2 = mysteryText2 + " ";
				}
				lastWasCJK = 1;
				lastWasSpace = 0;
				CJKcharacterAmount++;
			}
			else {
				if (lastWasCJK == 1 && mysteryChar != ' ') {
					mysteryText2 = mysteryText2 + " ";
				}
				if (mysteryChar == ' ') {
					lastWasSpace = 1;
				}
				else {
					lastWasSpace = 0;
				}
				lastWasCJK = 0;
			}
			mysteryText2 = mysteryText2 + mysteryChar;
		}
		
		mysteryText = mysteryText2;
		
		mysteryText = mysteryText.replaceAll("  *", " ");
		
		mysteryText = mysteryText.replaceAll("^ ", "");
		
		int strLength = mysteryText.length();
		
		if (strLength == 0) {
			return("xxx");
		}

		String[] words = mysteryText.split(" ");
        
        Map<String, Float> languagePoints = new HashMap();
		
		ListIterator languageIterator = languageList.listIterator();
		while(languageIterator.hasNext()) {
			Object element = languageIterator.next();
			String scoredLanguage = (String) element;
			languagePoints.put(scoredLanguage, (float) 0.0);
		}
		
		for (String mysteryWord : words) {
			Boolean wordScored = false;
			
			Map<String, Float> wordScores = new HashMap();
			
			mysteryWord = " " + mysteryWord + " ";
      
            TreeMap <String, Float> kiepro = new TreeMap<>();
			if (usedwords < 1) {
				if (wordDict.containsKey(mysteryWord)) {
					wordScored = true;
                    kiepro = wordDict.get(mysteryWord);
					languageIterator = languageList.listIterator();
					while(languageIterator.hasNext()) {
						Object element = languageIterator.next();
						String scoredLanguage = (String) element;
						if (kiepro.containsKey(scoredLanguage)) {
							wordScores.put(scoredLanguage, kiepro.get(scoredLanguage));
						}
						else {
							wordScores.put(scoredLanguage, penaltyValue);
						}
					}
				}
			}
			
			if (!wordScored) {
				languageIterator = languageList.listIterator();
				while(languageIterator.hasNext()) {
					Object element = languageIterator.next();
					String scoredLanguage = (String) element;
					wordScores.put(scoredLanguage, (float)  0.0);
				}
			}
			
			int t = maxNgram;
			while (t > 0) {
				if (wordScored) {
					break;
				}
				else {
					int pituus = mysteryWord.length();
					int x = 0;
					int grammaara = 0;
					if (pituus > (t-1)) {
						while (x < pituus - t + 1) {
							String gram = mysteryWord.substring(x,x+t);
							if (gramDict.containsKey(gram)) {
								grammaara = grammaara + 1;
								wordScored = true;
                                kiepro = gramDict.get(gram);
								languageIterator = languageList.listIterator();
								while(languageIterator.hasNext()) {
									Object element = languageIterator.next();
									String scoredLanguage = (String) element;
									if (kiepro.containsKey(scoredLanguage)) {
										wordScores.put(scoredLanguage, (wordScores.get(scoredLanguage)+kiepro.get(scoredLanguage)));
									}
									else {
										wordScores.put(scoredLanguage, (wordScores.get(scoredLanguage)+penaltyValue));
									}
								}
							}
							x = x + 1;
						}
					}
					if (wordScored) {
						languageIterator = languageList.listIterator();
						while(languageIterator.hasNext()) {
							Object element = languageIterator.next();
							String scoredLanguage = (String) element;
							wordScores.put(scoredLanguage, (wordScores.get(scoredLanguage)/grammaara));
						}
					}
				}
				t = t -1 ;
			}
			languageIterator = languageList.listIterator();
			while(languageIterator.hasNext()) {
				Object element = languageIterator.next();
				String scoredLanguage = (String) element;
				languagePoints.put(scoredLanguage, (languagePoints.get(scoredLanguage) + wordScores.get(scoredLanguage)));
			}
		}
		
		String winningLanguage = "xxx";
 		Float smallestScore = penaltyValue + 1;
        float wordNumber = words.length;
		
		languageIterator = languageList.listIterator();
		while(languageIterator.hasNext()) {
			Object element = languageIterator.next();
			String scoredLanguage = (String) element;
			languagePoints.put(scoredLanguage, (languagePoints.get(scoredLanguage)/wordNumber));
			if ((100/strLength*CJKcharacterAmount) > 50) {
				if (!scoredLanguage.equals("jpn") && !scoredLanguage.equals("kor") && !scoredLanguage.equals("cmn")) {
					languagePoints.put(scoredLanguage, (penaltyValue + 1));
				}
			}
			if (languagePoints.get(scoredLanguage) < smallestScore) {
				winningLanguage = scoredLanguage;
				smallestScore = languagePoints.get(element);
			}
		}
        
        String result = winningLanguage.substring(0,3);
        
        if (useConfidence) {
            Float secondScore = penaltyValue + 1;
            String secondLanguage = "xxx";
            
            languageIterator = languageList.listIterator();
            while(languageIterator.hasNext()) {
                Object element = languageIterator.next();
                String scoredLanguage = (String) element;
                if (scoredLanguage != winningLanguage) {
                    if (languagePoints.get(scoredLanguage) < secondScore) {
                        secondScore = languagePoints.get(scoredLanguage);
                        secondLanguage = scoredLanguage;
                    }
                }
            }

            Float confidence = secondScore - smallestScore;
            result = result + "\t" + confidence;
        }

		return (result);
	}
}

